/*
 * Copyright (c) 2023 Hiroo Ono <hiroo@oikumene.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
#include <sys/types.h>
#include <err.h>
#include <setjmp.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <png.h>

#define HBUFSIZE 32
#define MAGIC "SCRSHOT_"
#define DUMP_FMT_REV 2


void
usage(void)
{
	fprintf(stderr, "%s\n",
"vtscr infile outfile");
	exit(0);
}

void
read_header(FILE *fp, u_int *revision,
		u_int *width, u_int *height, u_int *depth)
{
	char header[HBUFSIZE];
	u_int32_t var;

	/* Check the header magic string */
	bzero(header, HBUFSIZE);
	if (fread(header, sizeof(u_char), strlen(MAGIC), fp)
			!= strlen(MAGIC))
		err(1, "fread() header bytes");
	if(strncmp(MAGIC, header, strlen(MAGIC)) != 0)
		err(1, "Expecting '%s', read '%s'", MAGIC, header);

	/* Check the revision of the dump format */
	header[0] = fgetc(fp);
	if (header[0] != DUMP_FMT_REV)
		err(1, "Expecting revision %d, read %d", DUMP_FMT_REV, header[0]);
	*revision = (u_int)header[0];

	/* Check and read the size data number */
	header[0] = fgetc(fp);
	if (header[0] != 3)
		err(1, "size data number is not 3 (%d)", (int)header[0]);
	if (fread(&var, sizeof(u_int32_t), 1, fp) != 1)
		err(1, "Could not read width");
	*width = var;
	if (fread(&var, sizeof(u_int32_t), 1, fp) != 1)
		err(1, "Could not read height");
	*height = var;
	if (fread(&var, sizeof(u_int32_t), 1, fp) != 1)
		err(1, "Could not read depth");
	*depth = var;
}

int
main(int argc, char** argv)
{
	FILE *ifp, *ofp;
	u_int32_t revision, width, height, depth, bpp;
	int bit_depth;
	u_int32_t *buffer;
	png_structp png_ptr;
	png_infop info_ptr;
	png_bytepp rows;
	int i, j, k;

	if (argc != 3)
		usage();
	ifp = fopen(argv[1], "rb");
	if (ifp == NULL)
		err(1, "Could not open file %s", argv[1]);

	read_header(ifp, &revision, &width, &height, &depth);
	switch (depth) {
		case 32:
			bit_depth = 8;
			bpp = depth / 8;
			break;
		default:
			err(1, "depth %u not supported", depth);
	}
	buffer = malloc(width * height * bpp);
	if (buffer == NULL) {
		fclose(ifp);
		err(1, "Could not allocate buffer");
	}
	if (fread(buffer, bpp, width*height, ifp) != width*height) {
		fclose(ifp);
		err(1, "fread() data bytes");
	}
	fclose(ifp);

	/* PNG data setup */
	png_ptr = png_create_write_struct(
			PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
	if (png_ptr == NULL)
		err(1, "setup failed");

	info_ptr = png_create_info_struct(png_ptr);
	if (info_ptr == NULL) {
		png_destroy_write_struct(&png_ptr, (png_infopp)NULL);
		err(1, "setup failed");
	}

	ofp = fopen(argv[2], "wb");
	if (ofp == NULL)
		err(1, "Could not open file %s", argv[2]);

	if (setjmp(png_jmpbuf(png_ptr))) {
		png_destroy_write_struct(&png_ptr, &info_ptr);
		fclose(ofp);
		exit(1);
	}

	png_init_io(png_ptr, ofp);
	png_set_IHDR(png_ptr, info_ptr, width, height, bit_depth,
			PNG_COLOR_TYPE_RGB,
			PNG_INTERLACE_NONE,
			PNG_COMPRESSION_TYPE_DEFAULT,
			PNG_FILTER_TYPE_DEFAULT);

	rows = png_malloc(png_ptr, sizeof(png_bytep) * height);
	png_set_rows(png_ptr, info_ptr, rows);
	for (i = 0; i < height; i++) {
		if (depth == 32) {
			rows[i] = png_malloc(png_ptr, width * 3);
			for (j = 0; j < width; j++) {
				memcpy(&rows[i][j*3], &buffer[i*width + j], 3);
			}
		}
		else {
			err(1, "depth (%u) not supported.", depth);
		}
	}

	png_write_png(png_ptr, info_ptr, PNG_TRANSFORM_BGR, NULL);
	fclose(ofp);
	png_destroy_write_struct(&png_ptr, &info_ptr);

	exit(0);
}
